@@ -5,7 +5,7 @@ use 5.014; use warnings;
 use Carp;
 use Scalar::Util qw< reftype blessed looks_like_number openhandle >;
 use overload;
-use Variable::Magic;
+use Sub::Uplevel;
 
 ### IF KEYWORDS {
 use Keyword::Declare;
@@ -102,7 +102,7 @@ sub _define_subtype {
             return sub {
                 my $okay = $old_type_handler->($_[0]);
                 return _error_near($_[0], $typename, $okay) if !$okay;
-                if (! eval{ $constraint->(local $_ = $_[0]) }) {
+                if (! eval{ local $SIG{__WARN__} = sub{}; $constraint->(local $_ = $_[0]) }) {
                     my $constraint_desc = _describe_constraint($_[0], undef, $constraint, $@);
                     return _error_near(
                         $_[0], qq{Value ($_[0]) did not satisfy the constraint: $constraint_desc\n }
@@ -114,7 +114,7 @@ sub _define_subtype {
     }
 
     $^H{"Dios::Types subtype=$new_typename"} = @user_defined_type;
-    push @user_defined_type, $new_type_handler_generator;
+    push @user_defined_type,            $new_type_handler_generator;
 }
 
 ### IF KEYWORDS }
@@ -139,6 +139,7 @@ sub _error_near ($$;$) {
 # Standard type checking...
 my %handler_for = (
     # Any Perl value or ref...
+    Slurpy     => sub { 1 },
     Any        => sub { 1 },
 
     # Anything that is true or false (and that's everything in Perl!)
@@ -147,6 +148,7 @@ my %handler_for = (
     # Anything defined, or not...
     Def        => sub {  defined $_[0] },
     Undef      => sub { !defined $_[0] },
+    Void       => sub { !defined $_[0] || ref $_[0] eq 'ARRAY' && !@{$_[0]} },
 
     # Values, references, and filehandles...
     Value      => sub { defined($_[0]) && !ref($_[0]) },
@@ -191,6 +193,25 @@ my %handler_for = (
         &looks_like_number
     },
 
+    # A string, or stringifiable object, or array ref, or hash ref, that is empty...
+    Empty => sub {
+        my $value = shift;
+
+        # Must be defined...
+        return 0 if !defined($value);
+
+        # May be an empty array or hash...
+        my $reftype = ref($value);
+        return 1 if $reftype eq 'ARRAY' && !@{$value};
+        return 1 if $reftype eq 'HASH'  && !keys %{$value};
+
+        # May be an object that overloads stringification...
+        return 1 if $reftype && overload::Method($value, q{""}) && "$value" eq q{};
+
+        # Otherwise, has to be an empty string...
+        return $value eq q{};
+    },
+
     # A string, or stringifiable object...
     Str => sub { defined($_[0]) && (ref($_[0]) ? overload::Method(shift,q{""}) : 1) },
 
@@ -221,7 +242,34 @@ for my $type (qw< SCALAR ARRAY HASH CODE GLOB >) {
     $handler_for{ ucfirst(lc($type)) } = sub { (reftype($_[0]) // q{}) eq $type };
 }
 $handler_for{ Regex } = sub { (reftype($_[0]) // q{}) eq 'REGEXP' };
+$handler_for{ List  } = $handler_for{ Array };
 
+# Standard type hierrachy...
+my %BASIC_NARROWER = (
+    Slurpy                           => {                                                               },
+        Any                          => { map {$_=>1} qw< Slurpy                                       >},
+            Bool                     => { map {$_=>1} qw< Slurpy Any                                   >},
+                Undef                => { map {$_=>1} qw< Slurpy Any Bool                              >},
+                Def                  => { map {$_=>1} qw< Slurpy Any Bool                              >},
+                    Value            => { map {$_=>1} qw< Slurpy Any Bool Def                          >},
+                        Num          => { map {$_=>1} qw< Slurpy Any Bool Def Value Str                >},
+                            Int      => { map {$_=>1} qw< Slurpy Any Bool Def Value Str Num            >},
+                        Str          => { map {$_=>1} qw< Slurpy Any Bool Def Value                    >},
+                            Class    => { map {$_=>1} qw< Slurpy Any Bool Def Value Str                >},
+                            Empty    => { map {$_=>1} qw< Slurpy Any Bool Def Value Str Ref Array List Hash >},
+                    Ref              => { map {$_=>1} qw< Slurpy Any Bool Def                          >},
+                        Scalar       => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        Regex        => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        Code         => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        Glob         => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        IO           => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        Obj          => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                        Array        => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                            List     => { map {$_=>1} qw< Slurpy Any Bool Def Ref Array                >},
+                            Empty    => { map {$_=>1} qw< Slurpy Any Bool Def Value Str Ref Array Hash List >},
+                        Hash         => { map {$_=>1} qw< Slurpy Any Bool Def Ref                      >},
+                            Empty    => { map {$_=>1} qw< Slurpy Any Bool Def Value Str Ref Array Hash List >},
+);
 
 # This is the full typename syntax...
 my $BASIC_TYPES = join('|', keys %handler_for);
@@ -246,11 +294,13 @@ my $TYPENAME_GRAMMAR = qr{
     |   Is    \[  (?<disj>   \s* (?&DISJ_TYPENAME)   \s* )  \]
     |   Is    \[  (?<conj>   \s* (?&CONJ_TYPENAME)   \s* )  \]
     |   Not   \[  (?<not>    \s* (?&CONJ_TYPENAME)   \s* )  \]
+    |   List  \[  (?<list>   \s* (?&CONJ_TYPENAME)   \s* )  \]
     |   Array \[  (?<array>  \s* (?&CONJ_TYPENAME)   \s* )  \]
     |   Tuple \[  (?<tuple>  \s* (?&TUPLE_FORMAT)    \s* )  \]
     |   Hash  \[  (?<hash>   \s* (?&CONJ_TYPENAME)   \s* )  \]
     |   Dict  \[  (?<dict>   \s* (?&DICT_FORMAT)     \s* )  \]
     |   Ref   \[  (?<ref>    \s* (?&CONJ_TYPENAME)   \s* )  \]
+    |   Eq    \[  (?<eq>     \s* (?&STR_SPEC)        \s* )  \]
     |   Match \[  (?<match>  \s* (?&REGEX_SPEC)      \s* )  \]
     |   Can   \[  (?<can>    \s* (?&OPT_QUAL_IDENT)  \s* )  \]
     |             (?<basic>      (?&BASIC)               )
@@ -283,7 +333,9 @@ my $TYPENAME_GRAMMAR = qr{
             $KEYED_TYPENAME
         )
 
-        (?<REGEX_SPEC> (?: [^][]++ |  \\[][]  |  \[ \^? \]? [^]]*+ \] )*+ )
+        (?<STR_SPEC>   (?: [^][\\]++ |  \\[][\\] )*+ )
+
+        (?<REGEX_SPEC> (?: [^][\\]++ |  \\\S  |  \[ \^? \]? [^]]*+ \] )*+ )
 
         (?<BASIC> \b (?: $BASIC_TYPES ) \b )
 
@@ -297,7 +349,8 @@ my $TYPENAME_GRAMMAR = qr{
 
 my $FROM_TYPENAME_GRAMMAR = qr{ (?(DEFINE) $TYPENAME_GRAMMAR ) }xms;
 
-my $IS_REF_TYPE = qr/\A (?: Array | Hash | Code | Scalar | Regex | Tuple | Dict | Glob | IO | Obj ) \b/x;
+my $IS_REF_TYPE
+    = qr/\A (?: List | Array | Hash | Code | Scalar | Regex | Tuple | Dict | Glob | IO | Obj ) \b/x;
 
 # Complex types are built on the fly...
 sub _build_handler_for {
@@ -380,6 +433,22 @@ sub _build_handler_for {
                                     }
                                   }
 
+    # List[T] types require an array ref, whose every element is of type T...
+    if ( exists $type_is{list} ) { my $value_handler = _build_handler_for($type_is{list});
+                                    return sub {
+                                        return _error_near($_[0], "List[$type_is{list}]")
+                                            if (reftype($_[0]) // q{}) ne 'ARRAY';
+
+                                        for (@{$_[0]}) {
+                                            my $okay = $value_handler->($_);
+                                            return _error_near($_, $type_is{list}, $okay)
+                                                if !$okay;
+                                        }
+
+                                        return 1;
+                                    }
+                                  }
+
     if ( exists $type_is{tuple} ) { my @types
                                         = grep {defined}
                                                $type_is{tuple} =~ m{ ((?&CONJ_TYPENAME) | $TYPED_OR_PURE_ETC )
@@ -516,6 +585,16 @@ sub _build_handler_for {
                                     }
                                   }
 
+    # Eq[S] types require a stringifiable, that matches 'S'...
+    if ( exists $type_is{eq}    ) { my $str = eval "q[$type_is{eq}]";
+                                    return sub {
+                                        return 1 if defined $_[0]
+                                                 && (!blessed($_[0]) || overload::Method($_[0],q{""}))
+                                                 && eval{ "$_[0]" eq $str };
+                                        return _error_near($_[0], "Eq[$type_is{eq}]");
+                                    }
+                                  }
+
     # Match[R] types require a stringifiable, that matches /R/x...
     if ( exists $type_is{match} ) { my $regex = eval { qr{$type_is{match}}x };
                                     croak "Invalid regex syntax in Match[$type_is{match}]:\n $@" if $@;
@@ -556,7 +635,7 @@ sub validate {
         }
 
         # Anything else is part of the value description...
-        else {
+        elsif (defined $arg) {
             $value_desc .= $arg;
         }
     }
@@ -573,7 +652,9 @@ sub validate {
     # Either the type matches or we die...
     if (!$handler_for{$typename}($value)) {
         $value_desc = _complete_desc($value_desc, $value);
-        croak qq{\u$value_desc is not of type $typename};
+        croak qq{\u$value_desc}
+            . ($value_desc =~ /\s$/ ? q{} : q{ })
+            . qq{is not of type $typename};
     }
     return 1 if !@constraints;
 
@@ -582,31 +663,351 @@ sub validate {
         local $@;
 
         # If it fails to match...
-        if (! eval{ $test->(local $_ = $value) }) {
+        if (! eval{ local $SIG{__WARN__} = sub{}; $test->(local $_ = $value) }) {
             $value_desc = _complete_desc($value_desc, $value);
             my $constraint_desc = _describe_constraint($value, $value_desc, $test, $@);
-            croak qq{\u$value_desc did not satisfy the constraint: $constraint_desc\n }
+            croak qq{\u$value_desc}
+                . ($value_desc =~ /\s$/ ? q{} : q{ })
+                . qq{did not satisfy the constraint: $constraint_desc\n }
         }
     }
 
     return 1;
 }
 
+package Dios::Types::TypedArray {
+    our @CARP_NOT = ('Dios::Types');
+    sub TIEARRAY  { bless [$_[1]], $_[0] }
+    sub FETCHSIZE { @{$_[0]} - 1 }
+    sub STORESIZE { $#{$_[0]} = $_[1] + 1 }
+    sub STORE     { my ($type, $desc, @constraint) = @{$_[0][0]};
+                    Dios::Types::validate($type, $_[2], $desc, @constraint);
+                    $_[0]->[$_[1]+1] = $_[2];
+                    }
+    sub FETCH     { $_[0]->[$_[1]+1] }
+    sub CLEAR     { @{$_[0]} = $_[0][0] }
+    sub POP       { @{$_[0]} > 1 ? pop(@{$_[0]}) : undef }
+    sub PUSH      { my $o = shift; push(@{$o}, @_) }
+    sub SHIFT     { splice(@{$_[0]},1,1) }
+    sub UNSHIFT   { my $o = shift; splice(@$o,1,0,@_) }
+    sub EXISTS    { exists $_[0]->[$_[1]+1] }
+    sub DELETE    { delete $_[0]->[$_[1]+1] }
+    sub EXTEND    { }
+
+    sub SPLICE
+    {
+        my $ob  = shift;
+        my $sz  = @{$ob} - 1;
+        my $off = @_ ? shift : 0;
+        $off   += $sz if $off < 0;
+        my $len = @_ ? shift : $sz-$off;
+        return splice(@$ob,$off+1,$len,@_);
+    }
+}
+
+package Dios::Types::TypedHash {
+    our @CARP_NOT = ('Dios::Types');
+    sub TIEHASH  { bless [$_[1], {}], $_[0] }
+    sub STORE    { my ($type, $desc, @constraint) = @{$_[0][0]};
+                   Dios::Types::validate($type, $_[2], $desc, @constraint);
+                   $_[0][1]{$_[1]} = $_[2]
+                 }
+    sub FETCH    { $_[0][1]{$_[1]} }
+    sub FIRSTKEY { my $a = scalar keys %{$_[0][1]}; each %{$_[0][1]} }
+    sub NEXTKEY  { each %{$_[0][1]} }
+    sub EXISTS   { exists $_[0][1]{$_[1]} }
+    sub DELETE   { delete $_[0][1]{$_[1]} }
+    sub CLEAR    { %{$_[0][1]} = () }
+    sub SCALAR   { scalar %{$_[0][1]} }
+}
+
 sub _set_var_type {
-    my ($type, undef, $value_desc) = @_;
-    if (!eval{ require Variable::Magic }) {
-        croak 'Typed attributes require the Variable::Magic module, which could not be loaded';
+    my ($type, $varref, $value_desc, @constraint) = @_;
+    my $vartype = ref $varref;
+
+    if ($vartype ne 'ARRAY' && $vartype ne 'HASH') {
+        croak 'Typed attributes require the Variable::Magic module, which could not be loaded'
+            if !eval{ require Variable::Magic };
+
+        Variable::Magic::cast ${$varref}, Variable::Magic::wizard set => sub {
+            # Code around awkward Object::Insideout behaviour...
+            return if ((caller 3)[3]//"") eq 'Object::InsideOut::DESTROY';
+
+            # Code around more awkward Object::Insideout behaviour...
+            no warnings 'redefine';
+            local *croak = *confess{CODE};
+            return if eval { validate($type, ${$_[0]}, $value_desc, @constraint) };
+            die $@ =~ s{\s+at .*}{}r
+                =~ s{[\h\S]*Dios.*}{}gr
+                =~ s{.*\(eval .*}{}gr
+                =~ s{\s*[\h\S]*called at}{ at}r
+                =~ s{.*called at.*}{}gr;
+        };
+    }
+    elsif ($vartype eq 'ARRAY') {
+        tie @{$varref}, 'Dios::Types::TypedArray', [$type, $value_desc, @constraint];
+    }
+    elsif ($vartype eq 'HASH') {
+        tie %{$varref}, 'Dios::Types::TypedHash',  [$type, $value_desc, @constraint];
+    }
+    else {
+        die 'Internal error: argument to _set_var_type() must be scalar, array ref, or hash ref';
+    }
+}
+
+# Implement return-type checking...
+sub _validate_return_type {
+
+    # Type info is first arg (an arrayref), subroutine body is final arg (a sub ref)...
+    my ($name, $type, $where) = @{shift()};
+    $where //= sub{1};
+    my $function = pop;
+
+    # List return context...
+    if (wantarray) {
+        # Tidy up type...
+        $type =~ s{\A Void \| | \| Void \Z}{}xmsg;
+        my $void_warning = vec((caller 1)[9], $warnings::Offsets{'void'}, 1);
+        warn sprintf "Call to $name() not in void context at %s line %d\n", (caller 1)[1,2]
+            if $void_warning && $type eq 'Void';
+
+        # Execute the subroutine body in (apparently) the right context...
+        my @retvals = uplevel 2, $function, @_;
+
+        # Adapt the constraint to produce a more appropriate error message...
+        my $listwhere = sub {
+            for (@{shift()}) {
+                die _describe_constraint($_,undef,$where) if !$where->($_)
+            }
+            return 1;
+        };
+
+        # Validate the return values...
+        eval {
+            if (@retvals == 1) {
+                validate(
+                    $type, $retvals[0], $where,
+                    "Return value (" . (_perl(@retvals)=~s/^\(|\)$//gr) . ") of call to $name()\n"
+                );
+            }
+            else {
+                undef;
+            }
+        }
+        //
+        eval {
+            validate(
+                $type, \@retvals, $listwhere,
+                "List of return values (" . (_perl(@retvals)=~s/^\(|\)$//gr) . ") of call to $name()\n"
+            )
+        }
+
+        # ..or convert the error message to report from the correct line number...
+        // die $@ =~ s{\s*at \S+ line \d+.*}{sprintf "\nat %s line %d\n", (caller 1)[1,2]}ser;
+
+        # If the return values are valid, return them...
+        return @retvals;
     }
 
-    Variable::Magic::cast $_[1], Variable::Magic::wizard set => sub {
-        # Code around awkward Object::Insideout behaviour...
-        return if ((caller 3)[3]//"") eq 'Object::InsideOut::DESTROY';
-#        return if ${^GLOBAL_PHASE} =~ /DESTRUCT|END/;
+    # Scalar context...
+    elsif (defined wantarray) {
+        # Tidy up type...
+        $type =~ s{\A Void \| | \| Void \Z}{}xmsg;
+        my $void_warning = vec((caller 1)[9], $warnings::Offsets{'void'}, 1);
+        warn sprintf "Call to $name() not in void context at %s line %d\n", (caller 1)[1,2]
+            if $void_warning && $type eq 'Void';
 
-        validate($type, ${$_[0]}, $value_desc);
-    };
+        # Execute the subroutine body in (apparently) the right context...
+        my $retval = uplevel 2, $function, @_;
+
+        # Validate the return value...
+        eval {
+            validate(
+                $type, $retval, $where,
+                "Scalar return value (" . _perl($retval) . ") of call to $name()\n"
+            )
+        }
+        # ...or convert the error message to report from the correct line number...
+        // die $@ =~ s{\s*at \S+ line \d+.*}{sprintf "\nat %s line %d\n", (caller 1)[1,2]}er;
+
+        # If the return value is valid, return it...
+        return $retval;
+    }
+
+    # Void context...
+    else {
+        # Execute the subroutine body in (apparently) the right context...
+        uplevel 2, $function, @_;
+
+        # Warn about explicit return types in void context, unless return type implies void is okay...
+        my $void_warning = vec((caller 1)[9], $warnings::Offsets{'void'}, 1);
+        warn sprintf
+            "Useless call to $name() with explicit return type $type\nin void context at %s line %d\n",
+            (caller 1)[1,2]
+                if $void_warning && !eval{ validate($type, undef) };
+
+    }
 }
 
+
+
+
+# Compare two types...
+sub _is_narrower {
+    my ($type_a, $type_b, $unnormalized) = @_;
+
+    # Short-circuit on identity...
+    return 0 if $type_a eq $type_b;
+
+    # Otherwise, normalize and decompose...
+    if (!$unnormalized && $type_a =~ m{\A (?: Ref ) \Z }xms) {
+        $type_a = "Ref[Any]";
+    }
+    elsif (!$unnormalized && $type_a =~ m{\A (?: Array | List ) \Z }xms) {
+        $type_a = "Ref[Array[Any]]";
+    }
+    elsif (!$unnormalized && $type_a eq 'Hash') {
+        $type_a = "Ref[Hash[Any]]";
+    }
+    elsif ($type_a =~ m{\A \s* ((?&NON_ATOM_TYPENAME)) \s* \Z  $FROM_TYPENAME_GRAMMAR }xms) {
+        $type_a = "Is[$1]";
+    }
+    $type_a =~ m{\A \s* $TYPENAME_GRAMMAR \s* \Z }xms;  my %type_a_is = %+;
+
+    if (!$unnormalized && $type_b =~ m{\A (?: Ref ) \Z }xms) {
+        $type_b = "Ref[Any]";
+    }
+    elsif (!$unnormalized && $type_b =~ m{\A (?: Array | List ) \Z }xms) {
+        $type_b = "Ref[Array[Any]]";
+    }
+    elsif (!$unnormalized && $type_b eq 'Hash') {
+        $type_b = "Ref[Hash[Any]]";
+    }
+    elsif ($type_b =~ m{\A \s* ((?&NON_ATOM_TYPENAME)) \s* \Z  $FROM_TYPENAME_GRAMMAR }xms) {
+        $type_b = "Is[$1]";
+    }
+    $type_b =~ m{\A \s* $TYPENAME_GRAMMAR \s* \Z }xms;  my %type_b_is = %+;
+
+    # If both are basic types, use the standard comparisons...
+    if (exists $type_a_is{basic} && exists $type_b_is{basic}) {
+        return +1 if $BASIC_NARROWER{$type_b}->{$type_a};
+        return -1 if $BASIC_NARROWER{$type_a}->{$type_b};
+    }
+
+    # If both are array or hash or reference types, use the standard comparisons on their element-types...
+    for my $elem_type (qw< array hash ref >) {
+        if (exists $type_a_is{$elem_type} && exists $type_b_is{$elem_type}) {
+            return _is_narrower($type_a_is{$elem_type}, $type_b_is{$elem_type}, 'unnormalized');
+        }
+    }
+
+    # If either type is parameterized, try the generic unparameterized version...
+    if ($type_a =~ s{\A(?:List|Array|Hash|Ref|Match|Eq)\K\[.*}{}xms
+    ||  $type_b =~ s{\A(?:List|Array|Hash|Ref|Match|Eq)\K\[.*}{}xms) {
+        return -1 if $type_a =~ m{\A(?:Match|Eq)\Z} && $BASIC_NARROWER{Class}->{$type_b};
+        return +1 if $type_b =~ m{\A(?:Match|Eq)\Z} && $BASIC_NARROWER{Class}->{$type_a};
+        return _is_narrower($type_a, $type_b, 'unnormalized');
+    }
+
+    # If both are user-defined types, try the standard inheritance hierarchy rules...
+    if (exists $type_a_is{user} && exists $type_b_is{user}) {
+        return +1 if $type_b->isa($type_a);
+        return -1 if $type_a->isa($type_b);
+    }
+
+    # Otherwise, unable to compare...
+    return 0;
+}
+
+# Compare two type signatures (of equal length)...
+sub _cmp_signatures {
+    my ($sig_a, $sig_b) = @_;
+
+    # Extract named parameters of B...
+    state %named_B_for;
+    my $named_B =
+        $named_B_for{$sig_b} //= { map { $_->{named} ? ($_->{named} => $_) : () } @{$sig_b} };
+
+    # Track relative ordering parameter-by-parameter...
+    my $partial_ordering = 0;
+    for my $n (0 .. max($#$sig_a, $#$sig_b)) {
+        # Unpack the next parameter types...
+        my $sig_a_n = $sig_a->[$n] // {};
+        my $sig_a_name = $sig_a_n->{named};
+        my $sig_b_n = ($sig_a_name ? $named_B->{$sig_a_name} : $sig_b->[$n]) // {};
+        my ($type_a, $type_b) = ($sig_a_n->{type} // 'Any', $sig_b_n->{type} // 'Any');
+
+        # Find the ordering of the next parameter pair from the two signatures...
+        my $is_narrower = _is_narrower($type_a, $type_b);
+
+        # Tie-break in favour of the type with more constraints...
+        if (!$is_narrower && $type_a eq $type_b) {
+            my $where_a = $sig_a_n->{where} // 0;
+            my $where_b = $sig_b_n->{where} // 0;
+            $is_narrower = $where_a > $where_b ? -1
+                         : $where_a < $where_b ? +1
+                         :                        0;
+        }
+
+        # If this pair's ordering contradicts the ordering so far, there is no ordering...
+        return 0 if $is_narrower && $is_narrower == -$partial_ordering;
+
+        # Otherwise if there's an ordering, it becomes the "ordering so far"...
+        $partial_ordering ||= $is_narrower;
+    }
+
+    # If we make it through the entire list, return the resulting ordering...
+    return $partial_ordering;
+}
+
+# Resolve ambiguous argument lists using Perl6-ish multiple dispatch rules...
+use List::Util qw< max first >;
+sub _resolve_signatures {
+    state %narrowness_for;
+    my ($kind, @sigs) = @_;
+
+    # Track narrownesses...
+    my %narrower = map { $_ => [] } 0..$#sigs;
+
+    # Compare all signatures, recording definitive differences in narrowness...
+    for my $index_1 (0 .. $#sigs) {
+        for my $index_2 ($index_1+1 .. $#sigs) {
+            my $sig1 = $sigs[$index_1]{sig};
+            my $sig2 = $sigs[$index_2]{sig};
+            my $narrowness =
+                $narrowness_for{$sig1,$sig2} //= _cmp_signatures($sig1, $sig2);
+
+            if    ($narrowness < 0) { push @{$narrower{$index_1}}, $index_2; }
+            elsif ($narrowness > 0) { push @{$narrower{$index_2}}, $index_1; }
+        }
+    }
+
+    # Find the narrowest signature(s)...
+    my $max_narrower = max map { scalar @{$_} } values %narrower;
+
+    # If they're not sufficiently narrow, weed out the non-contenders...
+    if ($max_narrower < @sigs-1) {
+        @sigs = @sigs[ sort grep { @{$narrower{$_}} } keys %narrower ];
+    }
+    # Otherwise, locate the narrowest...
+    else {
+        @sigs = @sigs[ first { @{$narrower{$_}} >= $max_narrower } keys %narrower ];
+    }
+
+    # Tie-break methods on the class of the variants...
+    if ($kind eq 'method' && @sigs > 1) {
+        @sigs = sort { $a->{class} eq $b->{class}    ?  0
+                     : $a->{class}->isa($b->{class}) ? -1
+                     : $b->{class}->isa($a->{class}) ? +1
+                     :                                  0
+                     } @sigs;
+        @sigs = grep { $_->{class} eq $sigs[0]{class} } @sigs;
+    }
+
+    return @sigs;
+}
+
+
 sub _describe_constraint {
     my ($value, $value_desc, $constraint, $constraint_desc) = @_;
 
@@ -636,23 +1037,18 @@ sub _describe_constraint {
         );
         $constraint_desc = $deparser->coderef2text($constraint);
         $constraint_desc =~ s{\s* BEGIN \s* \{ (?&CODE) \}
-                                (?(DEFINE) (?<CODE> [^{}]*+ (\{ (?&CODE) \} [^{}]*+ )*+ ))}{}xms;
-        $constraint_desc =~ s{(?: use \s*+ feature | die \s*+ sprintf ) [^;]* ;}{}gxms;
+                                (?(DEFINE) (?<CODE> [^{}]*+ (\{ (?&CODE) \} [^{}]*+ )*+ ))}{}gxms;
+        $constraint_desc =~ s{(?: (?:use|no) \s*+ feature | die \s*+ sprintf ) [^;]* ;}{}gxms;
         $constraint_desc =~ s{\s+}{ }g;
     }
     return $constraint_desc // "$constraint";
 }
 
 sub _perl {
-    my $value = shift;
+    use Data::Dump 'dump';
+    dump( map { my $classname = blessed $_; $classname ? "<$classname object>" : $_ } @_ )
+        =~ s{" (< \S+ \s object >) "}{$1}xgmsr;
 
-    if (my $classname = blessed $value) {
-        return "<$classname object>";
-    }
-    else {
-        use Data::Dump 'dump';
-        return dump($value);
-    }
 }
 
 
@@ -771,6 +1167,16 @@ converted to a string (e.g. a number), or any objects with a
 stringification overloading.
 
 
+=head3 C<< Empty >>
+
+Accepts any value that is a string, or a non-reference that can be
+converted to a string (e.g. a number), or any objects with a
+stringification overloading, provided the resulting string in
+each case is of zero length.
+
+Also accepts empty arrays and hashes (see below).
+
+
 =head3 C<< Class >>
 
 Accepts any value that's a string that is the name of a symbol-table
@@ -852,13 +1258,22 @@ the array must contain:
 
 Hence an unparameterized C<Array> is just a shorthand for C<Array[Any]>.
 
+The module also allows C<List> as a synonym for C<Array>.
+
+
+=head3 C<< Empty >>
+
+Accepts any value that is a reference to an array that contains no elements.
+
+Also accepts empty strings and hashes (see above and below).
+
 
 =head3 C<< Tuple[T1, T2, T3, ...], >>
 
 Accepts any value that is a reference to an array in which the sequence of
 array elements are of the specified types (in order). For example:
 
-    Array[Str, Int, Int, Hash]    # accepts: ["Foo", 1, 2,   {bar=>1}]
+    Tuple[Str, Int, Int, Hash]    # accepts: ["Foo", 1, 2,   {bar=>1}]
                                   # but not: ["Foo", 1, 2.1, {bar=>1}]
                                   # and not: [1, 2,  "Foo",  {bar=>1}]
 
@@ -866,7 +1281,7 @@ If the final specified type is followed by C<...>, the remainder of
 the elements may be any number of values (including none) of that
 type. For example:
 
-    Array[Str, Hash, Str...]  # accepts:  ["Foo", {bar=>1}]
+    Tuple[Str, Hash, Str...]  # accepts:  ["Foo", {bar=>1}]
                               # and also: ["Foo", {bar=>1}, 'cat']
                               # and also: ["Foo", {bar=>1}, 'cat', 'dog']
                               # et cetera...
@@ -874,7 +1289,7 @@ type. For example:
 If the last component of a tuple's type list is just C<...> by itself,
 the remainder of the elements may be anything (or nothing)...
 
-    Array[Str, Hash, ...]     # accepts:  ["Foo", {bar=>1}]
+    Tuple[Str, Hash, ...]     # accepts:  ["Foo", {bar=>1}]
                               # and also: ["Foo", {bar=>1}, 'etc']
                               # and also: ["Foo", {bar=>1}, 3, 4.5]
                               # et cetera...
@@ -895,6 +1310,13 @@ the hash may contain:
 
 Hence an unparameterized C<Hash> is just a shorthand for C<Hash[Any]>.
 
+=head3 C<< Empty >>
+
+Accepts any value that is a reference to an hash that contains no entries.
+
+Also accepts empty strings and arrays (see above).
+
+
 
 =head3 C<< Dict[ k, k => T, k? => T, ...], >>
 
@@ -943,15 +1365,26 @@ entry is present, there must also be a 'notes' array:
     Dict[name,ID]|Dict[name,ID,rank,notes=>Array]
 
 
+=head3 C<< Eq[STR] >>
+
+Accept a value whose stringification is C<eq> to C<'STR'>.
+
+The string is always assumed to be non-interpolating.
+
+Note that this type does not accept objects unless those objects
+overload stringification...even if the string specified would match the
+default C<'MyClass=HASH[0x1d15ed17]'> stringification of objects.
+
+
 =head3 C<< Match[PATTERN] >>
 
 Accept a value whose stringification matches the regex: C<m[PATTERN]x>
 
-The pattern is always assumed to have the C</x> modifier in effect. 
+The pattern is always assumed to have the C</x> modifier in effect.
 If you don't want that, you need to turn it off within the pattern:
 
     Match[      a b c ]     # accepts "abc"
-    Match[(?-x) a b c ]     # accepts " a b c"
+    Match[(?-x) a b c ]     # accepts " a b c "
 
 Note that this type does not accept objects unless those objects
 overload stringification...even if the pattern specified would match the
@@ -1098,10 +1531,11 @@ section form a single hierarchy, like so:
             \__Def
                 |__Value
                 |     |___Num
-                |     |    \__Int
+                |     |     \__Int
                 |     |
                 |      \__Str
-                |           \_Class
+                |          |___Empty
+                |           \__Class
                 |
                  \__Ref
                      |___Ref[<T>]
@@ -1112,10 +1546,12 @@ section form a single hierarchy, like so:
                      |___IO
                      |___Obj
                      |___Array
+                     |      |___Empty
                      |      |___Array[<T>]
                      |       \__Tuple[<T>, <T>, <T>, ...],
                      |
                       \__Hash
+                           |___Empty
                            |___Hash[<T>]
                             \__Dict[<k> => <T>, <k>? => <T>, ...],
 
@@ -1123,14 +1559,14 @@ That is, a value that is accepted by any specific type in this diagram
 will also be accepted by all of its ancestral types. So, for example,
 the type C<Tuple[Str,Int]> accepts the value S<C<['A',1]>>, so that
 same value will also be accepted by all of the following types (amongst
-many others): S<C<Tuple[Value,IntLike]>>, S<C<Tuple[Def,Num]>>,
+many others): S<C<Tuple[Value,Int]>>, S<C<Tuple[Def,Num]>>,
 S<C<Tuple[Any,Bool]>>, C<Array>, C<Ref>, C<Def>, C<Bool>, or C<Any>.
 
-Howwver, the converse is not generally true: a value that is accepted by
+However, the converse is not generally true: a value that is accepted by
 a "parent" type may not be accepted by all (or any) of its descendants.
 So while the type C<Array> accepts the value S<C<['A',{}]>>, that same
-value will B<not> be accepted by either of the "child" types:
-C<Array[Int]> or C<Tuple[Int,Str]>.
+value will B<not> be accepted by any of the "child" types:
+C<Empty>, C<Array[Int]>, or C<Tuple[Int,Str]>.
 
 
 =head1 INTERFACE
@@ -1355,7 +1791,7 @@ which is converted to a subroutine that is automatically passed to
 C<validate()> whenever the subtype is used.
 
 B<Note:> if you want to use the other features of this module, but
-don't want the compile-time overheads of this keyword, use the 
+don't want the compile-time overheads of this keyword, use the
 L<Dios::Types::Pure> module instead.
 
 =for IF KEYWORDS }
@@ -1442,7 +1878,11 @@ Dios::Types requires no configuration files or environment variables.
 
 Requires Perl 5.14 or later.
 
-Requires the Data::Dump and Variable::Magic modules.
+Requires the Data::Dump module.
+
+If typed attributes or parameters are used,
+also requires the Variable::Magic module.
+
 
 =for IF KEYWORDS {
 
